@@ -18,9 +18,8 @@ __docformat__ = 'restructuredtext'
 import persistent
 import transaction
 import zope.container.contained
-from urllib import urlencode
 
-from zope.interface import implements, Interface
+from zope.interface import implementer, Interface
 from zope.schema import TextLine
 from zope.publisher.interfaces.http import IHTTPRequest
 from zope.session.interfaces import ISession
@@ -29,6 +28,11 @@ from zope.traversing.browser.absoluteurl import absoluteURL
 from zope.site import hooks
 from zope.pluggableauth.interfaces import ICredentialsPlugin
 
+try:
+    from urllib import urlencode
+except ImportError:
+    # Py3: Location change.
+    from urllib.parse import urlencode
 
 class ISessionCredentials(Interface):
     """Interface for storing and accessing credentials in a session.
@@ -47,6 +51,7 @@ class ISessionCredentials(Interface):
         """Return password."""
 
 
+@implementer(ISessionCredentials)
 class SessionCredentials(object):
     """Credentials class for use with sessions.
 
@@ -64,7 +69,6 @@ class SessionCredentials(object):
       'tiger'
 
     """
-    implements(ISessionCredentials)
 
     def __init__(self, login, password):
         self.login = login
@@ -102,6 +106,7 @@ class IBrowserFormChallenger(Interface):
         default=u"password")
 
 
+@implementer(ICredentialsPlugin, IBrowserFormChallenger)
 class SessionCredentialsPlugin(persistent.Persistent,
                                zope.container.contained.Contained):
     """A credentials plugin that uses Zope sessions to get/store credentials.
@@ -129,31 +134,32 @@ class SessionCredentialsPlugin(persistent.Persistent,
       >>> from zope.pluggableauth.tests import sessionSetUp
       >>> from zope.publisher.browser import TestRequest
       >>> request = TestRequest()
-      >>> print plugin.extractCredentials(request)
+      >>> print(plugin.extractCredentials(request))
       None
 
     We must explicitly provide credentials once so the plugin can store
     them in a session:
 
       >>> request = TestRequest(login='scott', password='tiger')
-      >>> plugin.extractCredentials(request)
+      >>> from pprint import pprint
+      >>> pprint(plugin.extractCredentials(request))
       {'login': 'scott', 'password': 'tiger'}
 
     Subsequent requests now have access to the credentials even if they're
     not explicitly in the request:
 
-      >>> plugin.extractCredentials(TestRequest())
+      >>> pprint(plugin.extractCredentials(TestRequest()))
       {'login': 'scott', 'password': 'tiger'}
 
     We can always provide new credentials explicitly in the request:
 
-      >>> plugin.extractCredentials(TestRequest(
-      ...     login='harry', password='hirsch'))
+      >>> pprint(plugin.extractCredentials(TestRequest(
+      ...     login='harry', password='hirsch')))
       {'login': 'harry', 'password': 'hirsch'}
 
     and these will be used on subsequent requests:
 
-      >>> plugin.extractCredentials(TestRequest())
+      >>> pprint(plugin.extractCredentials(TestRequest()))
       {'login': 'harry', 'password': 'hirsch'}
 
     We can also change the fields from which the credentials are extracted:
@@ -168,14 +174,14 @@ class SessionCredentialsPlugin(persistent.Persistent,
 
     The plugin now extracts the credentials information from these new fields:
 
-      >>> plugin.extractCredentials(request)
+      >>> pprint(plugin.extractCredentials(request))
       {'login': 'luke', 'password': 'the_force'}
 
     Finally, we clear the session credentials using the logout method:
 
       >>> plugin.logout(TestRequest())
       True
-      >>> print plugin.extractCredentials(TestRequest())
+      >>> print(plugin.extractCredentials(TestRequest()))
       None
 
     Instances are persistent:
@@ -195,7 +201,6 @@ class SessionCredentialsPlugin(persistent.Persistent,
       True
 
     """
-    implements(ICredentialsPlugin, IBrowserFormChallenger)
 
     loginpagename = 'loginForm.html'
     loginfield = 'login'
@@ -317,19 +322,8 @@ class SessionCredentialsPlugin(persistent.Persistent,
             return False
 
         site = hooks.getSite()
-        # We need the traversal stack to complete the 'camefrom' parameter
-        stack = request.getTraversalStack()
-        stack.reverse()
-        # Better to add the query string, if present
-        query = request.get('QUERY_STRING')
-
-        camefrom = '/'.join([request.getURL()] + stack)
-        if query:
-            camefrom = camefrom + '?' + query
-        url = '%s/@@%s?%s' % (absoluteURL(site, request),
-                              self.loginpagename,
-                              urlencode({'camefrom': camefrom}))
-        request.response.redirect(url)
+        redirectWithComeFrom(request, '%s/@@%s' % (absoluteURL(site, request),
+                                                   self.loginpagename))
         return True
 
     def logout(self, request):
@@ -342,3 +336,53 @@ class SessionCredentialsPlugin(persistent.Persistent,
         sessionData['credentials'] = None
         transaction.commit()
         return True
+
+
+def redirectWithComeFrom(request, location):
+    """Redirect to a new location adding the current URL as ?comefrom=...
+
+          >>> from zope.publisher.browser import TestRequest
+          >>> request = TestRequest()
+
+          >>> redirectWithComeFrom(request, 'http://127.0.0.1/login')
+
+          >>> request.response.getStatus()
+          302
+          >>> request.response.getHeader('location')
+          'http://127.0.0.1/login?camefrom=http%3A%2F%2F127.0.0.1'
+
+    We'll fake up a more interesting request
+
+          >>> env = {
+          ...     'REQUEST_URI': '/foo/bar/folder/page%201.html?q=value',
+          ...     'QUERY_STRING': 'q=value'
+          ...     }
+          >>> request = TestRequest(environ=env)
+          >>> request._traversed_names = [u'foo', u'bar']
+          >>> request._traversal_stack = [u'page 1.html', u'folder']
+          >>> request['REQUEST_URI']
+          '/foo/bar/folder/page%201.html?q=value'
+
+          >>> redirectWithComeFrom(request, 'http://127.0.0.1/login')
+
+          >>> request.response.getHeader('location')
+          'http://127.0.0.1/login?camefrom=http%3A%2F%2F127.0.0.1%2Ffoo%2Fbar%2Ffolder%2Fpage+1.html%3Fq%3Dvalue'
+
+    """
+
+    # We need the traversal stack to complete the 'camefrom' parameter
+    # (sice this function must work during traversal as well as after it)
+    stack = request.getTraversalStack()
+    stack.reverse()
+
+    # Better to add the query string, if present
+    query = request.get('QUERY_STRING')
+
+    camefrom = '/'.join([request.getURL()] + stack)
+    if query:
+        camefrom = camefrom + '?' + query
+
+    # We assume location doesn't have query parameters
+    url = '%s?%s' % (location, urlencode({'camefrom': camefrom}))
+    request.response.redirect(url)
+
